/*
 Dual Output Linear Power Supply with Current Limiting
 Version: 0.1 Dev
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 16/03/2017
 Last Modified: 16/03/2017
 
 avrdude -c usbtiny -p atmega169a -U lfuse:w:0xef:m -U hfuse:w:0xd9:m -U efuse:w:0xff:m
 
 */

#define F_CPU 8000000 // 8 MHz

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/sleep.h>
#include "setup.c" // See defines, variables, constants, functions here

int main(void) {
	setup();
	
	//store_number(ps1SetVoltage, 1);
	//store_number(5678, 2);
	//store_number(98.76, 3);
	//store_number(1357, 4);
	
	uint8_t currentCounter = 0;
	
	while(1) {
		//display_led();
		//_delay_us(500);
		//_delay_ms(2);
		
		store_number(ps1Voltage, 1);
		store_number(ps1Current, 2);
		store_number(ps2Voltage, 3);
		store_number(ps2Current, 4);
		
		// Update stored value
		/*if (selectedDisplay == 1) {
			//ps1SetVoltage = selectedNumber;
			//ps1SetVoltage = ps1Voltage
		}
		else if (selectedDisplay == 2) {
			ps1SetCurrent = selectedNumber;
		}
		else if (selectedDisplay == 3) {
			ps2Voltage = selectedNumber;
		}
		else if (selectedDisplay == 4) {
			ps2Current = selectedNumber;
		}*/
		
		// left/right switch for changing displays
		/*if ((PING & (1<<LEFT_SWITCH)) || (PING & (1<<RIGHT_SWITCH))) {
			if (PING & (1<<LEFT_SWITCH)) {
				if (selectedDisplay > 1) {
					selectedDisplay--;
				}
			}
			else if (PING & (1<<RIGHT_SWITCH)) {
				if (selectedDisplay < 4) {
					selectedDisplay++;
				}
			}
			
			// Change number  
			if (selectedDisplay == 1) {
				selectedNumber = ps1SetVoltage;
			}
			else if (selectedDisplay == 2) {
				selectedNumber = ps1SetCurrent;
			}
			else if (selectedDisplay == 3) {
				selectedNumber = ps2Voltage;
			}
			else if (selectedDisplay == 4) {
				selectedNumber = ps2Current;
			}
			
			while (PING & (1<<LEFT_SWITCH));
			while (PING & (1<<RIGHT_SWITCH));
		}*/
		
		if ((PING & (1<<LEFT_SWITCH)) || (PING & (1<<RIGHT_SWITCH))) {
			if (PING & (1<<LEFT_SWITCH)) {
				if (selectedSegment > 0) {
					selectedSegment--;
				}
			}
			else if (PING & (1<<RIGHT_SWITCH)) {
				if (selectedSegment >= 0 && selectedSegment < 15) {
					selectedSegment++;
				}
			}
			
			if (selectedSegment == -1) {
				selectedSegment = previousSelectedSegment;
			}
			
			selectedDisplay = ((selectedSegment+4) / 4);
			
			if (selectedDisplay == 1) {
				selectedNumber = ps1Voltage;
			}
			else if (selectedDisplay == 2) {
				selectedNumber = ps1Current;
			}
			else if (selectedDisplay == 3) {
				selectedNumber = ps2Voltage;
			}
			else if (selectedDisplay == 4) {
				selectedNumber = ps2Current;
			}
			
			while (PING & (1<<LEFT_SWITCH));
			while (PING & (1<<RIGHT_SWITCH));
			_delay_ms(50);
		}
		
		if (selectedSegment >= 0 && ((PIND & (1<<PS1_EN_SWITCH)) || (PIND & (1<<PS2_EN_SWITCH))) ) {
			previousSelectedSegment = selectedSegment;
			selectedSegment = -1;
			while (PIND & (1<<PS1_EN_SWITCH));
			while (PIND & (1<<PS2_EN_SWITCH));
			_delay_ms(50);
			
			set_voltage();
		}
		
		// Calibrate ADC resistor divider if PS1 and PS2 enable are both pressed at the same time
		if ((PIND & (1<<PS1_EN_SWITCH)) && (PIND & (1<<PS2_EN_SWITCH))) {
			while (PIND & (1<<PS1_EN_SWITCH));
			while (PIND & (1<<PS2_EN_SWITCH));
			_delay_ms(50);
			
			inCalibration = true;
		}
		if (inCalibration == true) {
			ps2Current = adcResistorDivider;
			
			if ((PIND & (1<<PS1_EN_SWITCH)) || (PIND & (1<<PS2_EN_SWITCH))) {
				inCalibration = false;
				
				eeprom_write_byte((uint8_t*) ADC_RESISTOR_DIVIDER_EEPROM_LOCATION, adcResistorDivider >> 8);
				eeprom_write_byte((uint8_t*) ADC_RESISTOR_DIVIDER_EEPROM_LOCATION+1, adcResistorDivider & 0xFF);
			}
		}
		
		
		
		// Write PS1 LM350 DAC
		//soft_i2c_write_2_bytes_no_addr(PS1_VSET_DAC_ADDR, WRITE_DAC_REG | ((ps1SetVoltage & 0xF00) >> 8), (ps1SetVoltage & 0xFF), PS1_SCL, PS1_SDA);
		
		// Read PS1 LM350 ADC
		/*soft_i2c_read_2_bytes(PS1_VSENSE_ADC_ADDR, PS1_SCL, PS1_SDA);
		
		
		
		double adcresult = (double) ((i2cAdcHighByte << 8) | i2cAdcLowByte) * (double) 0.00006185 * (double) 11 * (double) 1000;
		ps2Voltage = (uint16_t) adcresult;*/
		
		
		
		
		
		soft_i2c_write_2_bytes_no_addr(PS1_VSET_DAC_ADDR, WRITE_DAC_REG | ((ps1LMDac & 0xF00) >> 8), (ps1LMDac & 0xFF), PS1_SCL, PS1_SDA);
		soft_i2c_write_2_bytes_no_addr(PS1_CSET_DAC_ADDR, WRITE_DAC_REG | ((ps1Current & 0xF00) >> 8), (ps1Current & 0xFF), PS1_SCL, PS1_SDA);
		
		_delay_ms(50);
		
		dcdc_regulate_voltage(10);
		
		if (inCalibration == false) {
			if (currentCounter >= 5) {
				uint16_t ps1CurrentAdcAvg = 0;
				for (uint8_t x = 0; x < 10; x++) {
					adc_start(PS1_CSENSE_ADC_PIN);
					_delay_ms(1);
					ps1CurrentAdcAvg += adc_result();
				}
				ps1CurrentAdcAvg = ps1CurrentAdcAvg / 10;
				
				ps2Current = (int) (double) ps1CurrentAdcAvg * (double) 0.00107421875 * (double) 2 * (double) 1000;
				
				currentCounter = 0;
			}
			currentCounter++;
		}
		
		if (selectedSegment == -1) {
			soft_i2c_read_2_bytes(PS1_VSENSE_ADC_ADDR, PS1_SCL, PS1_SDA);
			ps1Voltage = (uint16_t) (double) ((i2cAdcHighByte << 8) | i2cAdcLowByte) * (double) adcResolutionLsb * (double) (adcResistorDivider / (double) 1000) * (double) 1000;
			
			if (selectedDisplay == 2) {
				ps1Current = selectedNumber;
			}
		}
		else {
			//ps1Voltage = selectedNumber;
			
			if (selectedDisplay == 1) {
				ps1Voltage = selectedNumber;
			}
			else if (selectedDisplay == 2) {
				ps1Current = selectedNumber;
			}
			else if (selectedDisplay == 3) {
				ps2Voltage = selectedNumber;
			}
			else if (selectedDisplay == 4) {
				ps2Current = selectedNumber;
			}
		}
		
		//ps2Current = ps1LMDac;
		
		//_delay_ms(50);
	}
}

// Rotary encoder interrupt, set on the right pin, check if the left pin is high
ISR(INT0_vect) {
	if (PIND & (1<<ENCODER_L)) { // Left encoder is low, moving to the right
		if (selectedSegment >= 0) { // Selecting a segment to change
			uint8_t multiplier = selectedSegment - ((selectedDisplay*4) - 4);
			int16_t adjustment = 0;
			
			if (multiplier == 0) {
				if (selectedNumber < 10000) {
					adjustment = 1000;
				}
				else {
					adjustment = 10000;
				}
			}
			else if (multiplier == 1) {
				if (selectedNumber < 10000) {
					adjustment = 100;
				}
				else {
					adjustment = 1000;
				}
			}
			else if (multiplier == 2) {
				if (selectedNumber < 10000) {
					adjustment = 10;
				}
				else {
					adjustment = 100;
				}
			}
			else if (multiplier == 3) {
				if (selectedNumber < 10000) {
					adjustment = 1;
				}
				else {
					adjustment = 10;
				}
			}
			
			if (selectedDisplay == 1 || selectedDisplay == 3) {
				if ((selectedNumber + adjustment) < 15000) {
					selectedNumber += adjustment;
				}
			}
			else {
				if ((selectedNumber + adjustment) <= 2000) {
					selectedNumber += adjustment;
				}
			}
		}
		else if (inCalibration == true) {
			adcResistorDivider++;
		}
		else {
			if (selectedDisplay == 1 && ps1LMDac < 4095) {
				ps1LMDac += 1;
			}
			else if (selectedDisplay == 2 && selectedNumber < 2000) {
				selectedNumber += 1;
			}
		}
	}
	else {
		if (selectedSegment >= 0) {
			uint8_t multiplier = selectedSegment - ((selectedDisplay*4) - 4);
			int16_t adjustment = 0;
			
			if (multiplier == 0) {
				if ((selectedNumber / 10001) >= 1) {
					adjustment = -10000;
				}
				else {
					adjustment = -1000;
				}
			}
			else if (multiplier == 1) {
				if (selectedNumber <= 10000) {
					adjustment = -100;
				}
				else {
					adjustment = -1000;
				}
			}
			else if (multiplier == 2) {
				if (selectedNumber <= 10000) {
					adjustment = -10;
				}
				else {
					adjustment = -100;
				}
			}
			else if (multiplier == 3) {
				if (selectedNumber <= 10000) {
					adjustment = -1;
				}
				else {
					adjustment = -10;
				}
			}
			
			if ((selectedNumber + adjustment) > 0) {
				selectedNumber += adjustment;
			}
		}
		else if (inCalibration == true) {
			adcResistorDivider--;
		}
		else {
			if (selectedDisplay == 1 && ps1LMDac > 0) {
				ps1LMDac -= 1;
			}
			else if (selectedDisplay == 2 && selectedNumber > 0) {
				selectedNumber -= 1;
			}
		}
	}
}

// Timer0 overflow after 2ms
ISR(TIMER0_COMP_vect) {
	display_led();
}