/*
 Dual Output Linear Power Supply with Current Limiting
 Version: 0.1 Dev
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 16/03/2017
 Last Modified: 16/03/2017

 */

#ifndef cbi
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif

#ifndef NULL
#define NULL ((void *)0)
#endif

#define LOW 0
#define HIGH 1
#define false 0
#define true 1

#define LED_PORT_C 1
#define LED_PORT_B 0

#define I2C_DELAY_USEC 4
#define I2C_READ 1
#define I2C_WRITE 0


#define LEFT_SWITCH PG3
#define RIGHT_SWITCH PG4
#define ENCODER_L PD0
#define ENCODER_R PD1
#define PS1_EN_SWITCH PD2
#define PS2_EN_SWITCH PD3

#define WRITE_DAC_REG 0  // 00 command bits, 00 power down bits (Not Powered Down)
#define WRITE_DAC_CONFIG_ENABLE_VREF 0x98  // 100 command bits, 11 VRef, 00 power down bits (Not Powered Down), 0 Gain

#define PS1_VSET_DAC_ADDR 0x60
#define PS1_VSENSE_ADC_ADDR 0x68
#define PS1_DCDC_VSET_POT_ADDR 0x2F
#define PS1_CSET_DAC_ADDR 0x61

#define PS1_DCDC_VSENSE_ADC_PIN 7
#define PS1_CSENSE_ADC_PIN 6

#define PS1_SCL PF0
#define PS1_SDA PF1

// Minimum and max for the DC-DC module on both PS
#define DCDC_POT_MIN 20 // (
#define DCDC_POT_MAX 51 // (5V out)

// Pins to LCD display mapping
#define NA 0
#define NB 1
#define NC 2
#define ND 3
#define NE 4
#define NF 5
#define NG 6
#define NDP 7

#define DS1_SEG_COM1 PA0
#define DS1_SEG_COM2 PA1
#define DS1_SEG_COM3 PA2
#define DS1_SEG_COM4 PA3
#define DS2_SEG_COM1 PA4
#define DS2_SEG_COM2 PA5
#define DS2_SEG_COM3 PA6
#define DS2_SEG_COM4 PA7

#define DS3_SEG_COM1 PE0
#define DS3_SEG_COM2 PE1
#define DS3_SEG_COM3 PE2
#define DS3_SEG_COM4 PE3
#define DS4_SEG_COM1 PE4
#define DS4_SEG_COM2 PE5
#define DS4_SEG_COM3 PE6
#define DS4_SEG_COM4 PE7

// Number to led additional index
#define BLANK 10
#define LETTER_C 11
#define LETTER_A 12
#define LETTER_L 13
#define LETTER_D 14
#define LETTER_O 0
#define LETTER_N 15
#define LETTER_E 16
#define LETTER_R 17
#define LETTER_S 5
#define LETTER_T 18

volatile uint8_t ledDisplay[16] = {0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0}; // Storage of the number to display
volatile uint8_t segmentUpto = 0; // Which segment to light up
volatile uint16_t blinkDisplayCounter = 0;

uint8_t settingVoltage = 0;

int8_t previousSelectedSegment = 0;
int8_t selectedSegment = -1;

uint8_t selectedDisplay = 1;
volatile int16_t selectedNumber = 0;

uint16_t ps1Voltage = 0;
uint16_t ps1Current = 100;
uint16_t ps2Voltage = 3543;
uint16_t ps2Current = 1251;

uint16_t ps1LMDac = 0;
uint16_t ps1DcDcAdcVoltage = 0;
uint8_t ps1DcDcSettlingCount = 0;
uint8_t ps1DcDcVoltDPot = 51; // Default at power on

uint8_t i2cAdcHighByte = 0;
uint8_t i2cAdcLowByte = 0; 

double adcResolutionLsb = 0.0000625;
uint16_t adcResistorDivider = 7600;
#define ADC_RESISTOR_DIVIDER_EEPROM_LOCATION 1
uint8_t inCalibration = false;


// Mapping of characters to LEDs 
uint8_t number_to_led[19] = {
	(1<<NA) | (1<<NF) | (1<<NE) | (1<<ND) | (1<<NC) | (1<<NB), // 0
	(1<<NB) | (1<<NC), // 1
	(1<<NA) | (1<<NB) | (1<<NG) | (1<<NE) | (1<<ND), // 2
	(1<<NA) | (1<<NB) | (1<<NG) | (1<<NC) | (1<<ND), // 3
	(1<<NF) | (1<<NG) | (1<<NB) | (1<<NC), // 4
	(1<<NA) | (1<<NF) | (1<<NG) | (1<<NC) | (1<<ND), // 5
	(1<<NA) | (1<<NF) | (1<<NG) | (1<<NE) | (1<<ND) | (1<<NC), // 6
	(1<<NA) | (1<<NB) | (1<<NC), // 7
	(1<<NA) | (1<<NB) | (1<<NC) | (1<<ND) | (1<<NE) | (1<<NF) | (1<<NG), // 8
	(1<<NA) | (1<<NF) | (1<<NG) | (1<<NB) | (1<<NC), // 9
	0, // 10 (Blank)
	(1<<NA) | (1<<NF) | (1<<NE) | (1<<ND), // C
	(1<<NA) | (1<<NF) | (1<<NG) | (1<<NB) | (1<<NE) | (1<<NC), // A
	(1<<NF) | (1<<NE) | (1<<ND), // L
	(1<<NG) | (1<<NE) | (1<<ND) | (1<<NC) | (1<<NB), // D
	(1<<NE) | (1<<NG) | (1<<NC), // N
	(1<<NA) | (1<<NF) | (1<<NG) | (1<<NE) | (1<<ND), // E
	(1<<NE) | (1<<NG), // R
	(1<<NF) | (1<<NE) | (1<<NG) | (1<<ND), // T
};

// Light up the LEDs for the port, also includes blinking segments and deceminal points
void display_led_port (uint8_t port) {
	if (port == LED_PORT_C) {
		if (selectedSegment == segmentUpto && blinkDisplayCounter >= 100) {
			PORTC = number_to_led[(ledDisplay[segmentUpto])];
			if (PORTC == 0) { // Blinking segment is blank, put a 0 instead
				PORTC = number_to_led[0];
			}
		}
		else if (selectedSegment != segmentUpto) {
			PORTC = number_to_led[(ledDisplay[segmentUpto])];
		}
		if (settingVoltage == 0 && segmentUpto == 3 && selectedDisplay == 1 && blinkDisplayCounter >= 500) {
			PORTC |= (1<<NDP);
		}
		if (settingVoltage == 0 && segmentUpto == 7 && selectedDisplay == 2 && blinkDisplayCounter >= 500) {
			PORTC |= (1<<NDP);
		}
		if (segmentUpto == 0 && ps1Voltage < 10000) {
			PORTC |= (1<<NDP);
		}
		if (segmentUpto == 1 && ps1Voltage >= 10000) {
			PORTC |= (1<<NDP);
		}
	}
	else {
		if (selectedSegment == (segmentUpto+8) && blinkDisplayCounter >= 100) {
			PORTB = number_to_led[(ledDisplay[segmentUpto+8])];
		}
		else if (selectedSegment != (segmentUpto+8)) {
			PORTB = number_to_led[(ledDisplay[segmentUpto+8])];
		}
		if (settingVoltage == 0 && segmentUpto == 3 && selectedDisplay == 3 && blinkDisplayCounter >= 500) {
			PORTB |= (1<<NDP);
		}
		if (settingVoltage == 0 && segmentUpto == 7 && selectedDisplay == 4 && blinkDisplayCounter >= 500) {
			PORTB |= (1<<NDP);
		}
		if (segmentUpto == 0 && ps2Voltage < 10000) {
			PORTB |= (1<<NDP);
		}
		if (segmentUpto == 1 && ps2Voltage >= 10000) {
			PORTB |= (1<<NDP);
		}
	}
}

// Light up the LED segments two at a time (PS1 & PS2)
void display_led (void) {
	if (segmentUpto == 0) {
		PORTC = 0;
		PORTA |= (1<<DS2_SEG_COM4);
		PORTB = 0;
		PORTE |= (1<<DS4_SEG_COM4);
		
		PORTA &= ~(1<<DS1_SEG_COM1);
		display_led_port(LED_PORT_C);
		PORTE &= ~(1<<DS3_SEG_COM1);
		display_led_port(LED_PORT_B);
	}
	else if (segmentUpto == 1) {
		PORTC = 0;
		PORTA |= (1<<DS1_SEG_COM1);
		PORTB = 0;
		PORTE |= (1<<DS3_SEG_COM1);
		
		PORTA &= ~(1<<DS1_SEG_COM2);
		display_led_port(LED_PORT_C);
		PORTE &= ~(1<<DS3_SEG_COM2);
		display_led_port(LED_PORT_B);
	}
	else if (segmentUpto == 2) {
		PORTC = 0;
		PORTA |= (1<<DS1_SEG_COM2);
		PORTB = 0;
		PORTE |= (1<<DS3_SEG_COM2);
		
		PORTA &= ~(1<<DS1_SEG_COM3);
		display_led_port(LED_PORT_C);
		PORTE &= ~(1<<DS3_SEG_COM3);
		display_led_port(LED_PORT_B);
	}
	else if (segmentUpto == 3) {
		PORTC = 0;
		PORTA |= (1<<DS1_SEG_COM3);
		PORTB = 0;
		PORTE |= (1<<DS3_SEG_COM3);
		
		PORTA &= ~(1<<DS1_SEG_COM4);
		display_led_port(LED_PORT_C);
		PORTE &= ~(1<<DS3_SEG_COM4);
		display_led_port(LED_PORT_B);
	}
	else if (segmentUpto == 4) {
		PORTC = 0;
		PORTA |= (1<<DS1_SEG_COM4);
		PORTB = 0;
		PORTE |= (1<<DS3_SEG_COM4);
		
		PORTA &= ~(1<<DS2_SEG_COM1);
		display_led_port(LED_PORT_C);
		PORTE &= ~(1<<DS4_SEG_COM1);
		display_led_port(LED_PORT_B);
	}
	else if (segmentUpto == 5) {
		PORTC = 0;
		PORTA |= (1<<DS2_SEG_COM1);
		PORTB = 0;
		PORTE |= (1<<DS4_SEG_COM1);
		
		PORTA &= ~(1<<DS2_SEG_COM2);
		display_led_port(LED_PORT_C);
		PORTE &= ~(1<<DS4_SEG_COM2);
		display_led_port(LED_PORT_B);
	}
	else if (segmentUpto == 6) {
		PORTC = 0;
		PORTA |= (1<<DS2_SEG_COM2);
		PORTB = 0;
		PORTE |= (1<<DS4_SEG_COM2);
		
		PORTA &= ~(1<<DS2_SEG_COM3);
		display_led_port(LED_PORT_C);
		PORTE &= ~(1<<DS4_SEG_COM3);
		display_led_port(LED_PORT_B);
	}
	else if (segmentUpto == 7) {
		PORTC = 0;
		PORTA |= (1<<DS2_SEG_COM3);
		PORTB = 0;
		PORTE |= (1<<DS4_SEG_COM3);
		
		PORTA &= ~(1<<DS2_SEG_COM4);
		display_led_port(LED_PORT_C);
		PORTE &= ~(1<<DS4_SEG_COM4);
		display_led_port(LED_PORT_B);
		
		segmentUpto = 255; // Reset to zero on next increment
	}
	segmentUpto++;
	
	if (blinkDisplayCounter >= 2000) {
		blinkDisplayCounter = 0;
	}
	blinkDisplayCounter++;
}

// Store the number to the array
void store_number(uint16_t number, uint8_t displayNumber) {
	if (displayNumber == 1 || displayNumber == 3) { // Volts
		uint8_t element = 0;
		if (displayNumber == 3) {
			element = 8;
		}
		
		if (number >= 10000) {
			ledDisplay[element] = number / 10000;
			ledDisplay[element+1] = (number % 10000) / 1000;
			ledDisplay[element+2] = (number % 1000) / 100;
			ledDisplay[element+3] = (number % 100) / 10;
		}
		else {
			ledDisplay[element+0] = number / 1000;
			ledDisplay[element+1] = (number % 1000) / 100;
			ledDisplay[element+2] = (number % 100) / 10;
			ledDisplay[element+3] = number % 10;
		}
	}
	else if (displayNumber == 2 || displayNumber == 4) { // Amps
		uint8_t element = 4;
		if (displayNumber == 4) {
			element = 12;
		}
		
		ledDisplay[element] = BLANK;
		ledDisplay[element+1] = BLANK;
		ledDisplay[element+2] = BLANK;
		ledDisplay[element+3] = BLANK;
		
		int current = (int) number;
		
		if ((current / 1000) >= 1) {
			ledDisplay[element] = current / 1000;
			ledDisplay[element+1] = (current % 1000) / 100;
			ledDisplay[element+2] = (current % 100) / 10;
			ledDisplay[element+3] = current % 10;
		}
		else if ((current / 100) >= 1) {
			ledDisplay[element+1] = (current % 1000) / 100;
			ledDisplay[element+2] = (current % 100) / 10;
			ledDisplay[element+3] = current % 10;
		}
		else if ((current / 10) >= 1) {
			ledDisplay[element+2] = (current % 100) / 10;
			ledDisplay[element+3] = current % 10;
		}
		else {
			ledDisplay[element+3] = current % 10;
		}
	}
}


// Initialize SCL/SDA pins and set the bus high
void SoftI2cMasterInit(uint8_t TWI_SCL_PIN, uint8_t TWI_SDA_PIN) {
	DDRF |= (1<<TWI_SDA_PIN);
	PORTF |= (1<<TWI_SDA_PIN);
	DDRF |= (1<<TWI_SCL_PIN);
	PORTF |= (1<<TWI_SCL_PIN);
}

// De-initialize SCL/SDA pins and set the bus low
void SoftI2cMasterDeInit(uint8_t TWI_SCL_PIN, uint8_t TWI_SDA_PIN) {
	PORTF &= ~(1<<TWI_SDA_PIN);
	DDRF &= ~(1<<TWI_SDA_PIN);
	PORTF &= ~(1<<TWI_SCL_PIN);
	DDRF &= ~(1<<TWI_SCL_PIN);
}

// Read a byte from I2C and send Ack if more reads follow else Nak to terminate read
uint8_t SoftI2cMasterRead(uint8_t last, uint8_t TWI_SCL_PIN, uint8_t TWI_SDA_PIN) {
	uint8_t b = 0;
	// Make sure pull-up enabled
	PORTF |= (1<<TWI_SDA_PIN);
	DDRF &= ~(1<<TWI_SDA_PIN);
	// Read byte
	for (uint8_t i = 0; i < 8; i++) {
		// Don't change this loop unless you verify the change with a scope
		b <<= 1;
		_delay_us(I2C_DELAY_USEC);
		PORTF |= (1<<TWI_SCL_PIN);
		if (bit_is_set(PINF, TWI_SDA_PIN)) b |= 1;
		PORTF &= ~(1<<TWI_SCL_PIN);
	}
	// Send Ack or Nak
	DDRF |= (1<<TWI_SDA_PIN);
	if (last) { 
		PORTF |= (1<<TWI_SDA_PIN); 
	}
	else { 
		PORTF &= ~(1<<TWI_SDA_PIN);
	}  
	PORTF |= (1<<TWI_SCL_PIN);
	_delay_us(I2C_DELAY_USEC);
	PORTF &= ~(1<<TWI_SCL_PIN);
	PORTF &= ~(1<<TWI_SDA_PIN);
	return b;
}

// Write a byte to I2C
uint8_t SoftI2cMasterWrite(uint8_t data, uint8_t TWI_SCL_PIN, uint8_t TWI_SDA_PIN) {
	// Write byte
	for (uint8_t m = 0x80; m != 0; m >>= 1) {
		// Don't change this loop unless you verify the change with a scope
		if (m & data) { 
			PORTF |= (1<<TWI_SDA_PIN); 
		}
		else { 
			PORTF &= ~(1<<TWI_SDA_PIN); 
		}
		PORTF |= (1<<TWI_SCL_PIN);
		_delay_us(I2C_DELAY_USEC);
		PORTF &= ~(1<<TWI_SCL_PIN);
	}
	// get Ack or Nak
	DDRF &= ~(1<<TWI_SDA_PIN);
	// Enable pullup
	PORTF |= (1<<TWI_SDA_PIN);
	PORTF |= (1<<TWI_SCL_PIN);
	uint8_t rtn = bit_is_set(PINF, TWI_SDA_PIN);
	PORTF &= ~(1<<TWI_SCL_PIN);
	DDRF |= (1<<TWI_SDA_PIN);
	PORTF &= ~(1<<TWI_SDA_PIN);
	return rtn == 0;
}

// Issue a start condition
uint8_t SoftI2cMasterStart(uint8_t addressRW, uint8_t TWI_SCL_PIN, uint8_t TWI_SDA_PIN) {
	PORTF &= ~(1<<TWI_SDA_PIN);
	_delay_us(I2C_DELAY_USEC);
	PORTF &= ~(1<<TWI_SCL_PIN);
	return SoftI2cMasterWrite(addressRW, TWI_SCL_PIN, TWI_SDA_PIN);
}

// Issue a restart condition
uint8_t SoftI2cMasterRestart(uint8_t addressRW, uint8_t TWI_SCL_PIN, uint8_t TWI_SDA_PIN) {
	PORTF |= (1<<TWI_SDA_PIN);
	PORTF |= (1<<TWI_SCL_PIN);
	_delay_us(I2C_DELAY_USEC);
	return SoftI2cMasterStart(addressRW, TWI_SCL_PIN, TWI_SDA_PIN);
}

// Issue a stop condition
void SoftI2cMasterStop(uint8_t TWI_SCL_PIN, uint8_t TWI_SDA_PIN) {
	PORTF &= ~(1<<TWI_SDA_PIN);
	_delay_us(I2C_DELAY_USEC);
	PORTF |= (1<<TWI_SCL_PIN);
	_delay_us(I2C_DELAY_USEC);
	PORTF |= (1<<TWI_SDA_PIN);
	_delay_us(I2C_DELAY_USEC);
}

// Write 1 byte, no address
uint8_t soft_i2c_write_byte_no_addr(uint8_t deviceAddr, uint8_t writeByte, uint8_t TWI_SCL_PIN, uint8_t TWI_SDA_PIN) {
	
	// Issue a start condition, send device address and write direction bit
	if (!SoftI2cMasterStart((deviceAddr<<1) | I2C_WRITE, TWI_SCL_PIN, TWI_SDA_PIN)) return false;
	
	// Write the byte
	if (!SoftI2cMasterWrite(writeByte, TWI_SCL_PIN, TWI_SDA_PIN)) return false;

	// Issue a stop condition
	SoftI2cMasterStop(TWI_SCL_PIN, TWI_SDA_PIN);
	
	return true;
}

// Write 2 bytes, no address
uint8_t soft_i2c_write_2_bytes_no_addr(uint8_t deviceAddr, uint8_t writeByte1, uint8_t writeByte2, uint8_t TWI_SCL_PIN, uint8_t TWI_SDA_PIN) {
	
	// Issue a start condition, send device address and write direction bit
	if (!SoftI2cMasterStart((deviceAddr<<1) | I2C_WRITE, TWI_SCL_PIN, TWI_SDA_PIN)) return false;
	
	// Write the byte1
	if (!SoftI2cMasterWrite(writeByte1, TWI_SCL_PIN, TWI_SDA_PIN)) return false;
	
	// Write the byte2
	if (!SoftI2cMasterWrite(writeByte2, TWI_SCL_PIN, TWI_SDA_PIN)) return false;

	// Issue a stop condition
	SoftI2cMasterStop(TWI_SCL_PIN, TWI_SDA_PIN);
	
	return true;
}

// Read 2 bytes
uint8_t soft_i2c_read_2_bytes(uint8_t deviceAddr, uint8_t TWI_SCL_PIN, uint8_t TWI_SDA_PIN) {
	
	// Issue a start condition, send device address and write direction bit
	if (!SoftI2cMasterStart((deviceAddr<<1) | I2C_READ, TWI_SCL_PIN, TWI_SDA_PIN)) return false;
	
	i2cAdcHighByte = SoftI2cMasterRead(0, TWI_SCL_PIN, TWI_SDA_PIN);
	i2cAdcLowByte = SoftI2cMasterRead(1, TWI_SCL_PIN, TWI_SDA_PIN);
	
	// Issue a stop condition
	SoftI2cMasterStop(TWI_SCL_PIN, TWI_SDA_PIN);
	
	return true;
}

// Set the MCP4013 digial pot to a value (0 to 63)
/*void dpot_set(uint8_t value) {
	if (value != ps1DcDcVoltDPot) {
		if (value > ps1DcDcVoltDPot) {
			uint8_t changeAmount = value - ps1DcDcVoltDPot;
			
			PORTF |= (1<<PS1_UD_PIN);
			_delay_us(1);
			PORTF &= ~(1<<PS1_CS_PIN);
			_delay_us(2); // tlcuf/tlcur
			
			for (uint8_t x = 0; x < changeAmount; x++) {
				PORTF &= ~(1<<PS1_UD_PIN);
				_delay_us(2); // tlcur
				PORTF |= (1<<PS1_UD_PIN);
				_delay_us(1); // thi/fud
			}
			PORTF |= (1<<PS1_CS_PIN);
			
			ps1DcDcVoltDPot += changeAmount; // Update current value
		}
		else {
			uint8_t changeAmount = ps1DcDcVoltDPot - value;
			
			PORTF &= ~(1<<PS1_UD_PIN);
			_delay_us(1);
			PORTF &= ~(1<<PS1_CS_PIN);
			_delay_us(3); // tlcuf/tlcur
			PORTF |= (1<<PS1_UD_PIN); // Deincrement 1
			_delay_us(1); // tHI
			
			// Decrement more if amount is more than 1
			for (uint8_t x = 1; x < changeAmount; x++) {
				PORTF &= ~(1<<PS1_UD_PIN);
				_delay_us(1);
				PORTF |= (1<<PS1_UD_PIN);
				_delay_us(1); // tHI
			}
			PORTF |= (1<<PS1_CS_PIN);
			
			ps1DcDcVoltDPot -= changeAmount;
		}
	}
}*/


// Start the ADC with 1.1V reference
void adc_start(uint8_t pin) {
	// Set the analog reference (high two bits of ADMUX) and select the channel (low 4 bits)
	ADMUX = (1<<REFS1) | (1<<REFS0) | (pin & 0x3f); // 1.1V voltage reference
	
	sbi(ADCSRA, ADSC);
}

// Read ADC result
uint16_t adc_result(void) {
	// Read ADCL first
	uint8_t low = ADCL;
	uint8_t high = ADCH;
	
	// combine the two bytes
	return (high << 8) | low;
}

// Maintain the DC-DC output voltage at 2.5V to 3V above the LM350
// Can make it react quickly or slowly by adjusting the count, 5 = 250ms change time
void dcdc_regulate_voltage(uint8_t count) {
	ps1DcDcSettlingCount++;
	
	if (ps1DcDcSettlingCount >= count) {
		uint16_t ps1AdcAvg = 0;
		for (uint8_t x = 0; x < 10; x++) {
			adc_start(PS1_DCDC_VSENSE_ADC_PIN);
			_delay_ms(1);
			ps1AdcAvg += adc_result();
		}
		ps1DcDcAdcVoltage = ps1AdcAvg / 10;
		
		ps2Voltage = (int) (double) ps1DcDcAdcVoltage * (double) 0.00107421875 * (double) 21.06 * (double) 1000;
		
		int16_t voltageDiff = ps2Voltage - ps1Voltage;
		if (ps1DcDcVoltDPot > DCDC_POT_MIN && voltageDiff < 2500) {
			ps1DcDcVoltDPot--;
		}
		else if (ps1DcDcVoltDPot < DCDC_POT_MAX && voltageDiff > 3500){
			ps1DcDcVoltDPot++;
		}
		
		//dpot_set(ps1SetCurrent);
		soft_i2c_write_byte_no_addr(PS1_DCDC_VSET_POT_ADDR, ps1DcDcVoltDPot, PS1_SCL, PS1_SDA);
		
		ps1DcDcSettlingCount = 0;	
	}
}

void set_voltage(void) {//(uint8_t channel) {
	int16_t voltageSet = 0;
	settingVoltage = 1;
	
	//if (channel == 1) {
		voltageSet = ps1Voltage;
	//}
	
	
	while (settingVoltage == 1) {
		soft_i2c_write_2_bytes_no_addr(PS1_VSET_DAC_ADDR, WRITE_DAC_REG | ((ps1LMDac & 0xF00) >> 8), (ps1LMDac & 0xFF), PS1_SCL, PS1_SDA);
		_delay_ms(70);
		
		soft_i2c_read_2_bytes(PS1_VSENSE_ADC_ADDR, PS1_SCL, PS1_SDA);
		ps1Voltage = (uint16_t) (double) ((i2cAdcHighByte << 8) | i2cAdcLowByte) * (double) adcResolutionLsb * (double) (adcResistorDivider / (double) 1000) * (double) 1000;
		
		//adcResult = (uint16_t) (double) ((i2cAdcHighByte << 8) | i2cAdcLowByte) * (double) 0.00006185 * (double) 11 * (double) 1000;
		
		//ps2Voltage = adcResult;
		//ps2Current = ps1LMDac;
		
		
		int16_t difference = voltageSet - ps1Voltage;
		int16_t adjustment = 0;
		
		if (difference > 1000) {
			adjustment = 300;
			
			// Boost DC-DC voltage by a bit
			if (ps1DcDcVoltDPot > (DCDC_POT_MIN + 5)) {
				ps1DcDcVoltDPot-= 5;
			}
		}
		else if (difference < -1000) {
			adjustment = -300;
		}
		else if (difference > 500) {
			adjustment = 100;
		}
		else if (difference < -500) {
			adjustment = -100;
		}
		else if (difference > 100) {
			adjustment = 50;
		}
		else if (difference < -100) {
			adjustment = -50;
		}
		else if (difference > 20) {
			adjustment = 5;
		}
		else if (difference < -20) {
			adjustment = -5;
		}
		else if (difference > 2) {
			adjustment = 1;
		}
		else if (difference < -2) {
			adjustment = -1;
		}
		else {
			settingVoltage = 0;
		}
		
		// Check we can adjust the DAC to the level required, else exit
		if ((ps1LMDac + adjustment) > 0 && (ps1LMDac + adjustment) < 4095) {
			ps1LMDac += adjustment;
		}
		else {
			settingVoltage = 0;
		}
		
		
		dcdc_regulate_voltage(0);
		
		
		
		
		// coarse trim
		/*if ((voltageSet - adcResult) > 1000) {
			if (dacIncrement < 4000) {
				dacIncrement += 500;
			}
		}
		else if ((voltageSet - adcResult) < -1000) {
			if (dacIncrement > 50) {
				dacIncrement -= 500;
			}
		}
		
		// fine trim
		if ((voltageSet - adcResult) > 20) {
			
			if (dacIncrement < 4000) {
				dacIncrement += 5;
			}
		}
		else if ((voltageSet - adcResult) < -20) {
			
			if (dacIncrement > 50) {
				dacIncrement -= 5;
			}
		}
		else {
			finished = 1;
		}*/
		
		
		
		store_number(ps1Voltage, 1);
		store_number(ps1Current, 2);
		store_number(ps2Voltage, 3);
		store_number(ps2Current, 4);
		
		// Exit any button pressed
		if ((PIND & (1<<PS1_EN_SWITCH)) || (PIND & (1<<PS2_EN_SWITCH)) || 
			(PING & (1<<LEFT_SWITCH)) || (PING & (1<<RIGHT_SWITCH))) {
			settingVoltage = 0;
			
			while (PING & (1<<LEFT_SWITCH));
			while (PING & (1<<RIGHT_SWITCH));
			while (PIND & (1<<PS1_EN_SWITCH));
			while (PIND & (1<<PS2_EN_SWITCH));
			_delay_ms(50);
		}
		
		//_delay_ms(50);
	}
}


// Setup
void setup(void) {
	// Set outputs
	DDRA = 0xFF; // LED display 1
	DDRC = 0xFF;
	DDRE = 0xFF;
	DDRB = 0xFF; // LED display 4
	
	//DDRC |= (1<<PC0) | (1<<PC1) | (1<<PC2) | (1<<PC3) | (1<<PC4) | (1<<PC5);
	//DDRB |= (1<<PB1) | (1<<PB2);
	
	//PORTB |= (1<<latch);
	
	
	// Turn on ADC and set ADC prescale factor to 8
	// 8 MHz / 64 = 125 KHz, inside the desired 50-200 KHz range.
	ADCSRA = (1<<ADEN) | (1<<ADPS2) | (1<<ADPS1) | (1<<ADPS0);
	
	// Setup INT0 for rotary encoder
	EICRA = (1<<ISC01); // Falling edge
	sbi(EIMSK, INT0);
	
	_delay_ms(250);
	
	
	
	SoftI2cMasterInit(PS1_SCL, PS1_SDA);
	
	// Set ADC to 16bit
	soft_i2c_write_byte_no_addr(PS1_VSENSE_ADC_ADDR, 0x98, PS1_SCL, PS1_SDA);
	
	// Set VRef on DAC used for current limiting
	soft_i2c_write_byte_no_addr(PS1_CSET_DAC_ADDR, WRITE_DAC_CONFIG_ENABLE_VREF, PS1_SCL, PS1_SDA);
	
	// MCP4013 digital pot
	//DDRF |= (1<<PS1_UD_PIN) | (1<<PS1_CS_PIN);
	//PORTF |= (1<<PS1_UD_PIN) | (1<<PS1_CS_PIN); // Set high as idle
	
	// LED update timer
	TCNT0 = 0;
	TCCR0A = (1<<WGM01) | (1<<CS01) | (1<<CS00);
	sbi(TIMSK0, OCIE0A); // Enable interrupt
	OCR0A = 20;
	
	if (eeprom_read_byte((uint8_t*) ADC_RESISTOR_DIVIDER_EEPROM_LOCATION+1) != 255) {
		adcResistorDivider = (uint16_t) eeprom_read_byte((uint8_t*) ADC_RESISTOR_DIVIDER_EEPROM_LOCATION) << 8 | 
									(uint16_t) eeprom_read_byte((uint8_t*) ADC_RESISTOR_DIVIDER_EEPROM_LOCATION+1);
	}
	
	// Turn on interrupt
	sei();
}
